From e06b3a10c033955e14eff89abc8ed130f6de7adc Mon Sep 17 00:00:00 2001
From: Wu Jianhua <jianhua.wu@intel.com>
Date: Wed, 23 Feb 2022 20:03:03 +0800
Subject: [PATCH 06/10] avfilter/vf_blend_vulkan: add difference blend mode

Use the commands below to test: (href: https://trac.ffmpeg.org/wiki/Blend)

I. make an image for test
ffmpeg -f lavfi -i color=s=256x256,geq=r='H-1-Y':g='H-1-Y':b='H-1-Y' -frames 1 \
-y -pix_fmt yuv420p test.jpg

II. blend in sw
ffmpeg -i test.jpg -vf "split[a][b];[b]transpose[b];[a][b]blend=all_mode=difference,\
pseudocolor=preset=turbo" -y difference_sw.jpg

III. blend in vulkan
ffmpeg -init_hw_device vulkan -i test.jpg -vf "split[a][b];[b]transpose[b];\
[a]hwupload[a];[b]hwupload[b];[a][b]blend_vulkan=all_mode=difference,hwdownload,\
format=yuv420p,pseudocolor=preset=turbo" -y difference_vulkan.jpg

Signed-off-by: Wu Jianhua <jianhua.wu@intel.com>
---
 libavfilter/vf_blend_vulkan.c | 29 ++++++++++++++++-------------
 1 file changed, 16 insertions(+), 13 deletions(-)

diff --git a/libavfilter/vf_blend_vulkan.c b/libavfilter/vf_blend_vulkan.c
index b94081000e..9afdf79bfa 100644
--- a/libavfilter/vf_blend_vulkan.c
+++ b/libavfilter/vf_blend_vulkan.c
@@ -82,12 +82,13 @@ static const char blend_##MODE##_func[] = { \
 #define FN(EXPR) A + ((EXPR) - A) * opacity
 
 DEFINE_BLEND_MODE(NORMAL, A * opacity + B * (1.0f - opacity))
-DEFINE_BLEND_MODE(ADDITION,  FN(min(MAX, A + B)))
-DEFINE_BLEND_MODE(AVERAGE,   FN((A + B) * HALF))
-DEFINE_BLEND_MODE(SUBTRACT,  FN(max(MIN, A - B)))
-DEFINE_BLEND_MODE(MULTIPLY,  FN(1.0f * A * B / 1.0f))
-DEFINE_BLEND_MODE(NEGATION,  FN(MAX - abs(MAX - A - B)))
-DEFINE_BLEND_MODE(EXTREMITY, FN(abs(MAX - A - B)))
+DEFINE_BLEND_MODE(ADDITION,   FN(min(MAX, A + B)))
+DEFINE_BLEND_MODE(AVERAGE,    FN((A + B) * HALF))
+DEFINE_BLEND_MODE(SUBTRACT,   FN(max(MIN, A - B)))
+DEFINE_BLEND_MODE(MULTIPLY,   FN(1.0f * A * B / 1.0f))
+DEFINE_BLEND_MODE(NEGATION,   FN(MAX - abs(MAX - A - B)))
+DEFINE_BLEND_MODE(EXTREMITY,  FN(abs(MAX - A - B)))
+DEFINE_BLEND_MODE(DIFFERENCE, FN(abs(A - B)))
 
 static inline void init_blend_func(FilterParamsVulkan *param)
 {
@@ -104,6 +105,7 @@ static inline void init_blend_func(FilterParamsVulkan *param)
     CASE(MULTIPLY)
     CASE(NEGATION)
     CASE(EXTREMITY)
+    CASE(DIFFERENCE)
     default: param->blend = NULL; break;
     }
 
@@ -473,13 +475,14 @@ static const AVOption blend_vulkan_options[] = {
     { "c2_mode", "set component #2 blend mode", OFFSET(params[2].mode), AV_OPT_TYPE_INT, { .i64 = 0 }, 0, BLEND_NB - 1, FLAGS, "mode" },
     { "c3_mode", "set component #3 blend mode", OFFSET(params[3].mode), AV_OPT_TYPE_INT, { .i64 = 0 }, 0, BLEND_NB - 1, FLAGS, "mode" },
     { "all_mode", "set blend mode for all components", OFFSET(all_mode), AV_OPT_TYPE_INT, { .i64 = -1 }, -1, BLEND_NB - 1, FLAGS, "mode" },
-        { "addition",  "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_ADDITION  }, 0, 0, FLAGS, "mode" },
-        { "average",   "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_AVERAGE   }, 0, 0, FLAGS, "mode" },
-        { "subtract",  "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_SUBTRACT  }, 0, 0, FLAGS, "mode" },
-        { "normal",    "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_NORMAL    }, 0, 0, FLAGS, "mode" },
-        { "multiply",  "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_MULTIPLY  }, 0, 0, FLAGS, "mode" },
-        { "negation",  "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_NEGATION  }, 0, 0, FLAGS, "mode" },
-        { "extremity", "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_EXTREMITY }, 0, 0, FLAGS, "mode" },
+        { "addition",   "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_ADDITION   }, 0, 0, FLAGS, "mode" },
+        { "average",    "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_AVERAGE    }, 0, 0, FLAGS, "mode" },
+        { "subtract",   "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_SUBTRACT   }, 0, 0, FLAGS, "mode" },
+        { "normal",     "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_NORMAL     }, 0, 0, FLAGS, "mode" },
+        { "multiply",   "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_MULTIPLY   }, 0, 0, FLAGS, "mode" },
+        { "negation",   "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_NEGATION   }, 0, 0, FLAGS, "mode" },
+        { "extremity",  "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_EXTREMITY  }, 0, 0, FLAGS, "mode" },
+        { "difference", "", 0, AV_OPT_TYPE_CONST, { .i64 = BLEND_DIFFERENCE }, 0, 0, FLAGS, "mode" },
 
     { "c0_opacity",  "set color component #0 opacity", OFFSET(params[0].opacity), AV_OPT_TYPE_DOUBLE, { .dbl = 1 }, 0, 1, FLAGS },
     { "c1_opacity",  "set color component #1 opacity", OFFSET(params[1].opacity), AV_OPT_TYPE_DOUBLE, { .dbl = 1 }, 0, 1, FLAGS },
-- 
2.25.1

